package org.processmining.ptconversions.yawl;

import java.util.HashMap;
import java.util.Map;

import org.processmining.framework.util.Pair;
import org.processmining.processtree.Block;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.yawlfoundation.yawl.editor.elements.model.Decorator;
import org.yawlfoundation.yawl.editor.elements.model.EngineIdentifier;
import org.yawlfoundation.yawl.editor.elements.model.VertexContainer;
import org.yawlfoundation.yawl.editor.net.NetGraphModel;

public class YawlDefConverter implements YawlConverter {

	private YawlConversionType conversionType = YawlConversionType.PRIVATE;
	private Map<Node, NetGraphModel> subNets;
	private Map<NetGraphModel, Map<Node, Pair<VertexContainer, VertexContainer>>> fragments;

	public YawlDefConverter(YawlConversionType conversionType) {
		this.conversionType = conversionType;
		subNets = new HashMap<Node, NetGraphModel>();
		fragments = new HashMap<NetGraphModel, Map<Node, Pair<VertexContainer, VertexContainer>>>();
	}

	public Pair<VertexContainer, VertexContainer> convert(ProcessTree tree, Node node, NetGraphModel netGraphModel,
			VertexContainer exitNode, ProcessTree2Yawl pt2y, Map<Edge, Pair<EngineIdentifier, EngineIdentifier>> mappingEdgeIds) {
		if (conversionType == YawlConversionType.COMPOSITE && pt2y.hasMultipleParents(node)) {
			return convertComposite(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
		}
		if (conversionType == YawlConversionType.SHARED && pt2y.hasMultipleParents(node)) {
			return convertShared(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
		}
		if (conversionType == YawlConversionType.COMSHARED && pt2y.hasMultipleParents(node)) {
			return convertShared(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
		}
		if (conversionType == YawlConversionType.SIMPLE && pt2y.hasMultipleParents(node)) {
			return convertSimple(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
		}
		return convertPrivate(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
	}

	private Pair<VertexContainer, VertexContainer> convertPrivate(ProcessTree tree, Node node, NetGraphModel netGraphModel,
			VertexContainer exitNode, ProcessTree2Yawl pt2y, Map<Edge, Pair<EngineIdentifier, EngineIdentifier>> mappingEdgeIds) {
		
		VertexContainer alphaTask = YAWLUtil.createAtomicTask(netGraphModel, pt2y.newId());

		VertexContainer omegaTask = YAWLUtil.createAtomicTask(netGraphModel, pt2y.newId());
		YAWLUtil.setJoinDecorator(omegaTask, Decorator.XOR_TYPE);

		VertexContainer condition = YAWLUtil.createCondition(netGraphModel, "DEF" + pt2y.newId());
		
		YAWLUtil.createEdge(netGraphModel, alphaTask, condition);
		
		for (Node childNode : ((Block)node).getChildren()) {
			Pair<VertexContainer, VertexContainer> fragment = pt2y.convert(tree, childNode, netGraphModel, exitNode, pt2y, mappingEdgeIds);
			
			YAWLUtil.createEdge(netGraphModel, condition, fragment.getFirst());
			if (fragment.getSecond() != null) {
				YAWLUtil.createEdge(netGraphModel, fragment.getSecond(), omegaTask);
			}
		}
		
		return new Pair<VertexContainer, VertexContainer>(alphaTask, omegaTask);
	}
	
	private Pair<VertexContainer, VertexContainer> convertComposite(ProcessTree tree, Node node, NetGraphModel netGraphModel,
			VertexContainer exitNode, ProcessTree2Yawl pt2y, Map<Edge, Pair<EngineIdentifier, EngineIdentifier>> mappingEdgeIds) {
		
		NetGraphModel subNet;
		if (!subNets.containsKey(node)) {
			subNet = YAWLUtil.createSubnet(node, "DEF_net" + pt2y.newId());
			
			VertexContainer inputCondition = YAWLUtil.createInputCondition(subNet, pt2y.newId());
			VertexContainer outputCondition = YAWLUtil.createOutputCondition(subNet, pt2y.newId());
			
			for (Node childNode : ((Block)node).getChildren()) {
				Pair<VertexContainer, VertexContainer> fragment = pt2y.convert(tree, childNode, subNet, outputCondition, pt2y, mappingEdgeIds);
				Edge edge = ((Block)node).getOutgoingEdges().get(((Block)node).getChildren().indexOf(childNode));
				YAWLUtil.createEdge(subNet, inputCondition, fragment.getFirst(), edge.isHideable(), edge.isBlockable());
				mappingEdgeIds.put(edge, new Pair<EngineIdentifier, EngineIdentifier>(inputCondition.getVertex().getEngineIdentifier(), fragment.getFirst().getVertex().getEngineIdentifier()));
				if (fragment.getSecond() != null) {
					YAWLUtil.createEdge(subNet, fragment.getSecond(), outputCondition);
				}
			}
			
			subNets.put(node, subNet);
		} else {
			subNet = subNets.get(node);
		}
		
		VertexContainer compositeTask = YAWLUtil.createCompositeTask(netGraphModel, node.getName() + node.getID().toString());
		YAWLUtil.setDecomposition(compositeTask, subNet);
		YAWLUtil.setVariables(netGraphModel, compositeTask, node.getReadVariablesRecursive(), node.getWrittenVariablesRecursive(), false);
		
		return new Pair<VertexContainer, VertexContainer>(compositeTask, compositeTask);
	}

	private Pair<VertexContainer, VertexContainer> convertShared(ProcessTree tree, Node node, NetGraphModel netGraphModel,
			VertexContainer exitNode, ProcessTree2Yawl pt2y, Map<Edge, Pair<EngineIdentifier, EngineIdentifier>> mappingEdgeIds) {
		
		Pair<VertexContainer, VertexContainer> fragment;
		
		if (!fragments.containsKey(netGraphModel) || !fragments.get(netGraphModel).containsKey(node)) {
			
			VertexContainer inputCondition = YAWLUtil.createInputCondition(netGraphModel, pt2y.newId());
			VertexContainer outputCondition = YAWLUtil.createOutputCondition(netGraphModel, pt2y.newId());
			
			for (Node childNode : ((Block)node).getChildren()) {
				Pair<VertexContainer, VertexContainer> childFragment = pt2y.convert(tree, childNode, netGraphModel, outputCondition, pt2y, mappingEdgeIds);
				
				YAWLUtil.createEdge(netGraphModel, inputCondition, childFragment.getFirst());
				if (childFragment.getSecond() != null) {
					YAWLUtil.createEdge(netGraphModel, childFragment.getSecond(), outputCondition);
				}
			}
			
			fragment = new Pair<VertexContainer, VertexContainer>(inputCondition, outputCondition);
			if (!fragments.containsKey(netGraphModel)) {
				fragments.put(netGraphModel, new HashMap<Node, Pair<VertexContainer, VertexContainer>>());
			}
			fragments.get(netGraphModel).put(node, fragment);
		} 
			
		fragment = fragments.get(netGraphModel).get(node);
		
		VertexContainer alphaTask = YAWLUtil.createAtomicTask(netGraphModel, "DEF_alpha" + pt2y.newId());
		YAWLUtil.setSplitDecorator(alphaTask, Decorator.AND_TYPE);

		VertexContainer omegaTask = YAWLUtil.createAtomicTask(netGraphModel, "DEF_omega" + pt2y.newId());
		YAWLUtil.setJoinDecorator(omegaTask, Decorator.AND_TYPE);
				
		YAWLUtil.createEdge(netGraphModel, alphaTask, omegaTask);
		YAWLUtil.createEdge(netGraphModel, alphaTask, fragment.getFirst());
		YAWLUtil.createEdge(netGraphModel, fragment.getSecond(), omegaTask);
		
		return new Pair<VertexContainer, VertexContainer>(alphaTask, omegaTask);
	}

	private Pair<VertexContainer, VertexContainer> convertSimple(ProcessTree tree, Node node, NetGraphModel netGraphModel,
			VertexContainer exitNode, ProcessTree2Yawl pt2y, Map<Edge, Pair<EngineIdentifier, EngineIdentifier>> mappingEdgeIds) {
		
		Pair<VertexContainer, VertexContainer> fragment;
		
		if (!fragments.containsKey(netGraphModel) || !fragments.get(netGraphModel).containsKey(node)) {
			
			fragment = convertPrivate(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
			YAWLUtil.setJoinDecorator(fragment.getFirst(), Decorator.XOR_TYPE);
			YAWLUtil.setSplitDecorator(fragment.getSecond(), Decorator.XOR_TYPE);

			if (!fragments.containsKey(netGraphModel)) {
				fragments.put(netGraphModel, new HashMap<Node, Pair<VertexContainer, VertexContainer>>());
			}
			fragments.get(netGraphModel).put(node, fragment);
		} 
			
		fragment = fragments.get(netGraphModel).get(node);
		
		return fragment;
	}

}


